@@ -35,7 +35,6 @@ import torch.nn
 from .. import graph_break_hints, trace_rules, variables
 from ..exc import (
     raise_observed_exception,
-    unimplemented,
     unimplemented_v2,
     UnspecializeRestartAnalysis,
     Unsupported,
@@ -49,7 +48,6 @@ from ..source import (
     FSDPNNModuleSource,
     GetItemSource,
     NNModuleSource,
-    UnspecializedBuiltinNNModuleSource,
     UnspecializedNNModuleSource,
 )
 from ..utils import (
@@ -247,7 +245,17 @@ class NNModuleVariable(VariableTracker):
         base = tx.output.get_submodule(self.module_key)
 
         if object_has_getattribute(base):
-            unimplemented("NNModuleVariable with custom __getattribute__")
+            unimplemented_v2(
+                gb_type="Custom __getattribute__ in nn.Module dict key check",
+                context=f"has_key_in_generic_dict {self} {key}",
+                explanation="Dynamo does not support checking key existence "
+                "on `nn.Module` instances that have a custom "
+                "`__getattribute__` method defined.",
+                hints=[
+                    "Avoid defining `__getattribute__` in your module.",
+                    *graph_break_hints.SUPPORTABLE,
+                ],
+            )
 
         if tx.output.side_effects.has_pending_mutation_of_attr(self, key):
             mutated_attr = tx.output.side_effects.load_attr(self, key, deleted_ok=True)
@@ -259,14 +267,40 @@ class NNModuleVariable(VariableTracker):
     def _custom_getattr_fallback(self, base, tx, name, obj_source):
         """Check for a __getattr__ and handle it specially if it is implemented"""
         if object_has_getattribute(base):
-            unimplemented("torch.nn.Module with a custom __getattribute__ defined")
+            unimplemented_v2(
+                gb_type="Custom __getattribute__ in nn.Module attribute access",
+                context=f"var_getattr {self} {name}",
+                explanation="Dynamo does not support checking key existence "
+                "on `nn.Module` instances that have a custom "
+                "`__getattribute__` method defined.",
+                hints=[
+                    "Avoid defining `__getattribute__` in your module.",
+                    *graph_break_hints.SUPPORTABLE,
+                ],
+            )
 
         getattr_fn = get_custom_getattr(base, ignore_nn_module_getattr=True)
         if getattr_fn is None:
             return None
 
         if not isinstance(getattr_fn, types.FunctionType):
-            unimplemented("torch.nn.Module with a non-function custom __getattr__")
+            unimplemented_v2(
+                gb_type="torch.nn.Module with a non-function custom __getattr__",
+                context=f"var_getattr {self} {name}",
+                explanation=(
+                    "Dynamo detected a nn.Module object with a custom "
+                    "`__getattr__` method, but this method is not a standard "
+                    "Python function (e.g., it might be implemented in C/C++). "
+                    "Dynamo cannot currently trace into such non-standard "
+                    "`__getattr__` methods."
+                ),
+                hints=[
+                    "Avoid using objects with non-standard __getattr__ methods "
+                    "within the compiled region. If possible, implement "
+                    "__getattr__ as a standard Python function.",
+                    *graph_break_hints.SUPPORTABLE,
+                ],
+            )
 
         options = {"source": AttrSource(obj_source, "__getattr__")}
         return variables.UserMethodVariable(getattr_fn, self, **options).call_function(
@@ -284,7 +318,14 @@ class NNModuleVariable(VariableTracker):
             all_class_attribute_names.update(x.__dict__.keys())
 
         if not self.source:
-            unimplemented("GETATTR with no source")
+            unimplemented_v2(
+                gb_type="getattr with no source",
+                context=f"var_getattr {self} {name}",
+                explanation="Dynamo does not know how to access an attribute "
+                "on an `nn.Module` instance that lacks a source. This is "
+                "usually an internal error in Dynamo.",
+                hints=[*graph_break_hints.DYNAMO_BUG],
+            )
 
         if name == "__dict__":
             return variables.GetAttrVariable(self, name, source=source)
@@ -565,7 +606,13 @@ class NNModuleVariable(VariableTracker):
             if not all(
                 x.is_python_constant() for x in itertools.chain(args, kwargs.values())
             ):
-                unimplemented(f"non-const NNModule method {name}")
+                unimplemented_v2(
+                    gb_type="non-const argument in nn.Module method",
+                    context=f"call_method: {self} {name} {args} {kwargs}",
+                    explanation="Dynamo does not support calling "
+                    f"method `{name}` of ``nn.Module`` {module} with non-constant arguments.",
+                    hints=[],
+                )
 
         def get_kwargs(*names):
             assert_all_args_kwargs_const()
@@ -614,16 +661,14 @@ class NNModuleVariable(VariableTracker):
             return source
 
         if name == "named_children":
-            tx.output.guard_on_key_order.add(AttrSource(self.source, "_modules").name())
+            tx.output.guard_on_key_order.add(AttrSource(self.source, "_modules"))
             assert not (args or kwargs)
             result = []
             for name, submod in module.named_children():
                 result.append(named_embed(name, submod))
             return ListIteratorVariable(result, mutation_type=ValueMutationNew())
         elif name == "named_parameters":
-            tx.output.guard_on_key_order.add(
-                AttrSource(self.source, "_parameters").name()
-            )
+            tx.output.guard_on_key_order.add(AttrSource(self.source, "_parameters"))
             result = []
             for name, param in module.named_parameters(
                 **get_kwargs("prefix", "recurse")
@@ -631,7 +676,7 @@ class NNModuleVariable(VariableTracker):
                 result.append(named_embed(name, param))
             return ListIteratorVariable(result, mutation_type=ValueMutationNew())
         elif name == "named_buffers":
-            tx.output.guard_on_key_order.add(AttrSource(self.source, "_buffers").name())
+            tx.output.guard_on_key_order.add(AttrSource(self.source, "_buffers"))
             result = []
             for name, buffer in module.named_buffers(
                 **get_kwargs("prefix", "recurse", "remove_duplicate")
@@ -639,7 +684,7 @@ class NNModuleVariable(VariableTracker):
                 result.append(named_embed(name, buffer))
             return ListIteratorVariable(result, mutation_type=ValueMutationNew())
         elif name == "named_modules":
-            tx.output.guard_on_key_order.add(AttrSource(self.source, "_modules").name())
+            tx.output.guard_on_key_order.add(AttrSource(self.source, "_modules"))
             result = []
             for name, submod in module.named_modules(
                 **get_kwargs("memo", "prefix", "remove_duplicate")
@@ -647,19 +692,17 @@ class NNModuleVariable(VariableTracker):
                 result.append(named_embed(name, submod))
             return ListIteratorVariable(result, mutation_type=ValueMutationNew())
         elif name == "children":
-            tx.output.guard_on_key_order.add(AttrSource(self.source, "_modules").name())
+            tx.output.guard_on_key_order.add(AttrSource(self.source, "_modules"))
             assert not (args or kwargs)
             return wrap_values(module.named_children())
         elif name == "modules":
-            tx.output.guard_on_key_order.add(AttrSource(self.source, "_modules").name())
+            tx.output.guard_on_key_order.add(AttrSource(self.source, "_modules"))
             return wrap_values(module.named_modules())
         elif name == "parameters":
-            tx.output.guard_on_key_order.add(
-                AttrSource(self.source, "_parameters").name()
-            )
+            tx.output.guard_on_key_order.add(AttrSource(self.source, "_parameters"))
             return wrap_values(module.named_parameters(**get_kwargs("recurse")))
         elif name == "buffers":
-            tx.output.guard_on_key_order.add(AttrSource(self.source, "_buffers").name())
+            tx.output.guard_on_key_order.add(AttrSource(self.source, "_buffers"))
             return wrap_values(module.named_buffers(**get_kwargs("recurse")))
         elif name == "keys":
             assert not (args or kwargs)
@@ -756,7 +799,13 @@ class NNModuleVariable(VariableTracker):
             elif args[0].is_python_constant():
                 key = args[0].as_python_constant()
             else:
-                unimplemented(f"getitem on NNModuleVariable with key {args[0]}")
+                unimplemented_v2(
+                    gb_type="Unsupported key type for nn.Module.__getitem__",
+                    context=f"call_method: {self} {name} {args} {kwargs}",
+                    explanation="Dynamo does not support getitem on "
+                    "`nn.Module` with non-constant key.",
+                    hints=[],
+                )
 
             submod = module[key]
             return tx.output.register_attr_or_module(
@@ -819,7 +868,7 @@ class UnspecializedNNModuleVariable(UserDefinedObjectVariable):
         if type(value) is torch.jit._script.RecursiveScriptModule:
             raise Unsupported(
                 "ScriptModules aren't supported in UnspecializedNNModuleVariable"
-                " becuase their .forward function isn't a static member of their type"
+                " because their .forward function isn't a static member of their type"
             )
         if "value_type" in kwargs:
             lazy_value_to_become = getattr(kwargs["value_type"], "cls_to_become", None)
@@ -841,8 +890,7 @@ class UnspecializedNNModuleVariable(UserDefinedObjectVariable):
         self.nn_module_stack_source = self.source
 
     def _wrap_source(self, attr_source):
-        if not isinstance(attr_source, UnspecializedNNModuleSource):
-            return UnspecializedNNModuleSource(attr_source)
+        # the vt is already wrapped with UnspecializedNNModuleSource
         return attr_source
 
     def get_nn_module_stack_source(self):
@@ -852,7 +900,7 @@ class UnspecializedNNModuleVariable(UserDefinedObjectVariable):
         self.nn_module_stack_source = source
 
     @staticmethod
-    @functools.lru_cache(None)
+    @functools.cache
     def _nn_module_method_ids():
         # Allow __setattr__ to fall through to base class handler
         supported = {torch.nn.Module.__setattr__, torch.nn.Module.__init__}
@@ -991,14 +1039,22 @@ class UnspecializedNNModuleVariable(UserDefinedObjectVariable):
                 hasattr(method, "__code__")
                 and id(method.__code__) in self._nn_module_method_ids()
             ):
-                unimplemented(f"UnspecializedNNModuleVariable missing {name}")
+                unimplemented_v2(
+                    gb_type="UnspecializedNNModuleVariable missing method",
+                    context=f"call_method: {self} {name} {args} {kwargs}",
+                    explanation=f"Dynamo does not support tracing method {name} of nn.Module {self.value}",
+                    hints=[
+                        "Dynamo does not really define unspecialized nn.Module very well.",
+                        *graph_break_hints.DIFFICULT,
+                    ],
+                )
 
             # "_parameters" in self.value.__dict__ checks that module is initialized
             if name == "__setattr__" and "_parameters" in self.value.__dict__:
                 # Record if mutations happens on parameters/buffers/modules. The
                 # mutations on these are not tracked by base class
                 # UserDefinedObject vt. This will be used later to graph break
-                # on seeing a paramters() and family calls.
+                # on seeing a parameters() and family calls.
                 # TODO(anijain2305) - This might not be needed if we let Dynamo
                 # inline both getattr and setattr. In that case, it should see
                 # the lowest level dicts - _parameters and family and
@@ -1074,7 +1130,7 @@ class UnspecializedNNModuleVariable(UserDefinedObjectVariable):
 
         # For non-empty hook dicts, one way is to just fallback to VariableTracker.build() and create a ConstDictVariable.
         # However, ConstDictVariable guards on keys. This can cause recompiles when the same hook is installed for
-        # differnt nn module instances, because the key keeps changing (look more into RemovableHandle to understand why
+        # different nn module instances, because the key keeps changing (look more into RemovableHandle to understand why
         # key changes - also related https://github.com/pytorch/pytorch/issues/125836). Here, we carefully craft a
         # NNModuleHooksDictVariable (a subclass of ConstDictVariable) to avoid any guard on the keys.
         if (
@@ -1089,7 +1145,7 @@ class UnspecializedNNModuleVariable(UserDefinedObjectVariable):
             hooks_dict = getattr(self.value, name)
             hooks_dict_source = AttrSource(self.source, name)
             install_guard(hooks_dict_source.make_guard(GuardBuilder.SEQUENCE_LENGTH))
-            tx.output.guard_on_key_order.add(hooks_dict_source.name())
+            tx.output.guard_on_key_order.add(hooks_dict_source)
 
             def build_key_value(i, k, v):
                 # Make key sourceless to avoid any guard on it
@@ -1135,8 +1191,7 @@ class UnspecializedBuiltinNNModuleVariable(UnspecializedNNModuleVariable):
     """
 
     def _wrap_source(self, attr_source):
-        if not isinstance(attr_source, UnspecializedBuiltinNNModuleSource):
-            return UnspecializedBuiltinNNModuleSource(attr_source)
+        # vt is already wrapped with the UnspecializedBuiltinNNModuleSource
         return attr_source
 
 
